from __future__ import annotations

from typing import TYPE_CHECKING, Self

from pydantic import Field, JsonValue, model_validator
from pydantic_zarr.v2 import AnyArraySpec, AnyGroupSpec, GroupSpec

from ome_zarr_models._utils import _from_zarr_v2
from ome_zarr_models.base import BaseAttrsv2
from ome_zarr_models.common.coordinate_transformations import _build_transforms
from ome_zarr_models.v04.axes import Axis
from ome_zarr_models.v04.base import BaseGroupv04
from ome_zarr_models.v04.labels import Labels
from ome_zarr_models.v04.multiscales import Dataset, Multiscale
from ome_zarr_models.v04.omero import Omero

if TYPE_CHECKING:
    from collections.abc import Sequence

    import zarr


__all__ = ["Image", "ImageAttrs"]


class ImageAttrs(BaseAttrsv2):
    """
    Metadata for OME-Zarr image groups.
    """

    multiscales: list[Multiscale] = Field(
        ...,
        description="The multiscale datasets for this image",
        min_length=1,
    )
    omero: Omero | None = None

    def get_array_paths(self) -> list[str]:
        paths = []
        for multiscale in self.multiscales:
            for dataset in multiscale.datasets:
                paths.append(dataset.path)
        return paths

    def get_optional_group_paths(self) -> dict[str, type[AnyGroupSpec]]:
        return {"labels": Labels}


class Image(BaseGroupv04[ImageAttrs]):
    """
    An OME-Zarr multiscale dataset.
    """

    @classmethod
    def from_zarr(cls, group: zarr.Group) -> Self:  # type: ignore[override]
        """
        Create an OME-Zarr image model from a `zarr.Group`.

        Parameters
        ----------
        group : zarr.Group
            A Zarr group that has valid OME-Zarr image metadata.
        """
        return _from_zarr_v2(group, cls, ImageAttrs)

    @classmethod
    def new(
        cls,
        *,
        array_specs: Sequence[AnyArraySpec],
        paths: Sequence[str],
        axes: Sequence[Axis],
        scales: Sequence[Sequence[float]],
        translations: Sequence[Sequence[float] | None],
        name: str | None = None,
        multiscale_type: str | None = None,
        metadata: JsonValue | None = None,
        global_scale: Sequence[float] | None = None,
        global_translation: Sequence[float] | None = None,
    ) -> Image:
        """
        Create a new `Image` from a sequence of multiscale arrays
        and spatial metadata.

        Parameters
        ----------
        array_specs :
            A sequence of array specifications that collectively represent the same
            image at multiple levels of detail.
        paths :
            The paths to the arrays within the new Zarr group.
        axes :
            `Axis` objects describing the axes of the arrays.
        scales :
            For each array, a scale value for each axis of the array.
        translations :
            For each array, a translation value for each axis the array.
        name :
            A name for the multiscale collection.
        multiscale_type :
            Type of downscaling method used to generate the multiscale image pyramid.
            Optional.
        metadata :
            Arbitrary metadata to store in the multiscales group.
        global_scale :
            A global scale value for each axis of every array.
        global_translation :
            A global translation value for each axis of every array.

        Notes
        -----
        This class does not store or copy any array data. To save array data,
        first write this class to a Zarr store, and then write data to the Zarr
        arrays in that store.
        """
        if len(array_specs) != len(paths):
            raise ValueError(
                f"Length of arrays (got {len(array_specs)=}) must be the same as "
                f"length of paths (got {len(paths)=})"
            )
        members_flat: dict[str, AnyArraySpec] = {
            "/" + key.lstrip("/"): arr
            for key, arr in zip(paths, array_specs, strict=True)
        }

        if global_scale is None and global_translation is None:
            global_transform = None
        elif global_scale is None:
            raise ValueError(
                "If global_translation is specified, "
                "global_scale must also be specified."
            )
        else:
            global_transform = _build_transforms(global_scale, global_translation)

        if len(scales) != len(paths):
            raise ValueError(
                f"Length of 'scales' ({len(scales)}) does not match "
                f"length of 'paths' {len(paths)}"
            )
        if len(translations) != len(paths):
            raise ValueError(
                f"Length of 'translations' ({len(translations)}) does not match "
                f"length of 'paths' ({len(paths)})"
            )
        multimeta = Multiscale(
            axes=tuple(axes),
            datasets=tuple(
                Dataset.build(path=path, scale=scale, translation=translation)
                for path, scale, translation in zip(
                    paths, scales, translations, strict=True
                )
            ),
            coordinateTransformations=global_transform,
            metadata=metadata,
            name=name,
            type=multiscale_type,
            version="0.4",
        )
        return Image(
            members=GroupSpec.from_flat(members_flat).members,
            attributes=ImageAttrs(multiscales=(multimeta,)),
        )

    @model_validator(mode="after")
    def _check_arrays_compatible(self) -> Self:
        """
        Check that all the arrays referenced by the `multiscales` metadata meet the
        following criteria:
            - they exist
            - they are not groups
            - they have dimensionality consistent with the number of axes defined in the
              metadata.
        """
        multimeta = self.attributes.multiscales
        flat_self = self.to_flat()

        for multiscale in multimeta:
            multiscale_ndim = len(multiscale.axes)
            for dataset in multiscale.datasets:
                try:
                    maybe_arr: AnyArraySpec | AnyGroupSpec = flat_self[
                        "/" + dataset.path.lstrip("/")
                    ]
                    if isinstance(maybe_arr, GroupSpec):
                        msg = f"The node at {dataset.path} is a group, not an array."
                        raise ValueError(msg)
                    arr_ndim = len(maybe_arr.shape)

                    if arr_ndim != multiscale_ndim:
                        msg = (
                            f"The multiscale metadata has {multiscale_ndim} axes "
                            "which does not match the dimensionality of the array "
                            f"found in this group at {dataset.path} ({arr_ndim}). "
                            "The number of axes must match the array dimensionality."
                        )

                        raise ValueError(msg)
                except KeyError as e:
                    msg = (
                        f"The multiscale metadata references an array that does not "
                        f"exist in this group: {dataset.path}"
                    )
                    raise ValueError(msg) from e
        return self

    @property
    def labels(self) -> Labels | None:
        """
        Labels group contained within this image group.

        Returns `None` if no labels are present.

        Raises
        ------
        RuntimeError
            If the node at "labels" is not a group.
        """
        if self.members is None or "labels" not in self.members:
            return None

        labels_group = self.members["labels"]
        if not isinstance(labels_group, GroupSpec):
            raise RuntimeError("Node at 'labels' is not a group")

        return Labels(attributes=labels_group.attributes, members=labels_group.members)

    @property
    def datasets(self) -> tuple[tuple[Dataset, ...], ...]:
        """
        Get datasets stored in this image.

        The first index is for the multiscales.
        The second index is for the dataset inside that multiscales.
        """
        return tuple(
            tuple(dataset for dataset in multiscale.datasets)
            for multiscale in self.attributes.multiscales
        )



from collections.abc import Generator, Mapping
from typing import TYPE_CHECKING, Self

# Import needed for pydantic type resolution
import pydantic_zarr  # noqa: F401
import zarr
from pydantic import model_validator
from pydantic_zarr.v3 import GroupSpec

from ome_zarr_models._utils import _from_zarr_v3
from ome_zarr_models.common.well import WellGroupNotFoundError
from ome_zarr_models.v05.base import BaseGroupv05, BaseOMEAttrs
from ome_zarr_models.v05.plate import Plate
from ome_zarr_models.v05.well import Well

if TYPE_CHECKING:
    from pydantic_zarr.v3 import AnyGroupSpec


__all__ = ["HCS", "HCSAttrs"]


class HCSAttrs(BaseOMEAttrs):
    """
    HCS metadtata attributes.
    """

    plate: Plate

    def get_optional_group_paths(self) -> dict[str, type[Well]]:  # type: ignore[override]
        return {well.path: Well for well in self.plate.wells}


class HCS(BaseGroupv05[HCSAttrs]):
    """
    An OME-Zarr high content screening (HCS) dataset.
    """

    @classmethod
    def from_zarr(cls, group: zarr.Group) -> Self:  # type: ignore[override]
        """
        Create an OME-Zarr image model from a `zarr.Group`.

        Parameters
        ----------
        group : zarr.Group
            A Zarr group that has valid OME-Zarr image metadata.
        """
        hcs = _from_zarr_v3(group, cls, HCSAttrs)
        # Traverse all the Well groups, which themselves contain Image groups
        hcs_flat = hcs.to_flat()
        for well in hcs.ome_attributes.plate.wells:
            if well.path in group:
                well_group = group[well.path]
                well_group_flat = Well.from_zarr(well_group).to_flat()  # type: ignore[arg-type]
                for path in well_group_flat:
                    hcs_flat["/" + well.path + path] = well_group_flat[path]

        hcs_unflat: AnyGroupSpec = GroupSpec.from_flat(hcs_flat)
        return cls(attributes=hcs_unflat.attributes, members=hcs_unflat.members)

    @model_validator(mode="after")
    def _check_valid_acquisitions(self) -> Self:
        """
        Check well acquisition IDs are in list of plate acquisition ids.
        """
        acquisitions = self.ome_attributes.plate.acquisitions
        if acquisitions is None:
            return self

        valid_aq_ids = [aq.id for aq in acquisitions]

        for well_i, well_group in enumerate(self.well_groups):
            for image_i, well_image in enumerate(well_group.ome_attributes.well.images):
                if well_image.acquisition is None:
                    continue
                elif well_image.acquisition not in valid_aq_ids:
                    msg = (
                        f"Acquisition ID '{well_image.acquisition} "
                        f"(found in well {well_i}, {image_i}) "
                        f"is not in list of plate acquisitions: {valid_aq_ids}"
                    )
                    raise ValueError(msg)

        return self

    @property
    def n_wells(self) -> int:
        """
        Number of wells.
        """
        return len(self.ome_attributes.plate.wells)

    @property
    def well_groups(self) -> Generator[Well, None, None]:
        """
        Well groups within this HCS group.

        Notes
        -----
        Only well groups that exist are returned. This can be less than the number
        of wells defined in the HCS metadata if some of the well Zarr groups don't
        exist.
        """
        for i in range(self.n_wells):
            try:
                yield self.get_well_group(i)
            except WellGroupNotFoundError:
                continue

    def get_well_group(self, i: int) -> Well:
        """
        Get a single well group.

        Parameters
        ----------
        i :
            Index of well group.

        Raises
        ------
        WellGroupNotFoundError
            If no Zarr group is found at the well path.
        """
        if self.members is None:
            raise RuntimeError("Zarr group has no members")

        well = self.ome_attributes.plate.wells[i]
        well_path = well.path
        well_path_parts = well_path.split("/")
        if len(well_path_parts) != 2:
            raise RuntimeError(f"Well path '{well_path_parts}' does not have two parts")
        row, col = well_path_parts
        if row not in self.members:
            raise WellGroupNotFoundError(
                f"Row '{row}' not found in group members: {self.members}"
            )
        if (
            not isinstance(row_group := self.members[row], GroupSpec)
            or not isinstance(row_group.members, Mapping)
            or col not in row_group.members
        ):
            raise WellGroupNotFoundError(
                f"Column '{col}' not found in row group members: {self.members[row]}"
            )
        group = row_group.members[col]
        return Well(attributes=group.attributes, members=group.members)



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Literal, Self, TypeVar, Union

import pydantic_zarr
import pydantic_zarr.v3
from pydantic import BaseModel

from ome_zarr_models.base import BaseAttrsv3, BaseGroup

if TYPE_CHECKING:
    import zarr


class BaseOMEAttrs(BaseAttrsv3):
    """
    Base class for OME-Zarr 0.6 attributes.
    """

    version: Literal["0.6"]


T = TypeVar("T", bound=BaseOMEAttrs)


class BaseZarrAttrs(BaseModel, Generic[T]):
    """
    Base class for zarr attributes in an OME-Zarr group.
    """

    ome: T


class BaseGroupv06(
    BaseGroup,
    pydantic_zarr.v3.GroupSpec[
        BaseZarrAttrs[T],
        Union["pydantic_zarr.v3.GroupSpec", "pydantic_zarr.v3.ArraySpec"],  # type: ignore[type-arg]
    ],
    Generic[T],
):
    """
    Base class for all v0.6 OME-Zarr groups.
    """

    @classmethod
    def from_zarr(cls, group: zarr.Group) -> Self:  # type: ignore[override]
        """
        Create an OME-Zarr model from a `zarr.Group`.

        Parameters
        ----------
        group : zarr.Group
            A Zarr group that has valid OME-Zarr image metadata.
        """
        return super().from_zarr(group)

    @property
    def ome_zarr_version(self) -> Literal["0.6"]:
        """
        OME-Zarr version.
        """
        return "0.6"

    @property
    def ome_attributes(self) -> T:
        """
        OME attributes.
        """
        return self.attributes.ome



from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Never, TypeVar

import pytest
import zarr
import zarr.storage
from zarr.abc.store import Store
from zarr.storage import LocalStore, MemoryStore

from ome_zarr_models.base import BaseAttrs

if TYPE_CHECKING:
    from zarr.abc.store import Store


T = TypeVar("T", bound=BaseAttrs)

EXAMPLES_PATH = Path(__file__).parent / "data" / "examples"
Version = Literal["0.4", "0.5", "0.6"]


def get_examples_path(*, version: Version) -> Path:
    """
    Get path to directory with example data.
    """
    if version == "0.4":
        return EXAMPLES_PATH / "v04"
    elif version == "0.5":
        return EXAMPLES_PATH / "v05"
    elif version == "0.6":
        return EXAMPLES_PATH / "v06"
    else:
        raise ValueError(f"Unknown value for {version=}")


def json_to_dict(*, version: Version, json_fname: str) -> Any:
    """
    Load an example JSON file and return as a dictionary.
    """
    with open(get_examples_path(version=version) / json_fname) as f:
        return json.load(f)


def read_in_json(*, version: Version, json_fname: str, model_cls: type[T]) -> T:
    """
    Load an example JSON file into a ome-zarr-models base attributes class.
    """
    with open(get_examples_path(version=version) / json_fname) as f:
        return model_cls.model_validate_json(f.read())


def json_to_zarr_group(
    *, version: Version, json_fname: str, store: Store
) -> zarr.Group:
    """
    Create an empty Zarr group, and set the attributes from an example JSON file.
    """
    zarr_format: Literal[2, 3]
    if version == "0.4":
        zarr_format = 2
    elif version in ["0.5", "0.6"]:
        zarr_format = 3
    else:
        raise ValueError(f"Unknown value for {version=}")
    group = zarr.open_group(store=store, zarr_format=zarr_format)
    with open(get_examples_path(version=version) / json_fname) as f:
        attrs = json.load(f)

    group.attrs.put(attrs)
    return group


class UnlistableStore(MemoryStore):
    """
    A memory store that doesn't support listing.

    Mimics other remote stores (e.g., HTTP) that don't support listing.
    """

    supports_listing: bool = False

    def list(self) -> Never:
        raise NotImplementedError

    def list_dir(self, prefix: str) -> Never:
        raise NotImplementedError

    def list_prefix(self, prefix: str) -> Never:
        raise NotImplementedError


@pytest.fixture(params=["MemoryStore", "LocalStore", "UnlistableStore"])
def store(request: pytest.FixtureRequest, tmp_path: Path) -> Store:
    match request.param:
        case "MemoryStore":
            return MemoryStore()
        case "LocalStore":
            return LocalStore(root=tmp_path)
        case "UnlistableStore":
            return UnlistableStore()
        case _:
            raise RuntimeError(f"Unknown store class: {request.param}")



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Self

from pydantic import Field

from ome_zarr_models.base import BaseAttrs

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

__all__ = [
    "Identity",
    "PathScale",
    "PathTranslation",
    "ScaleTransform",
    "TranslationTransform",
    "VectorScale",
    "VectorTransform",
    "VectorTranslation",
]


class Transform(BaseAttrs):
    type: Literal["identity", "scale", "translation"]


class Identity(Transform):
    """
    Identity transformation.

    Notes
    -----
    Although defined in the specification, it is not allowed
    to be used anywhere.
    """

    type: Literal["identity"]


class VectorScale(Transform):
    """
    Scale transformation parametrized by a vector of numbers.
    """

    type: Literal["scale"]
    scale: list[float]

    @classmethod
    def build(cls, data: Iterable[float]) -> Self:
        """
        Create a VectorScale from an iterable of floats.
        """
        return cls(type="scale", scale=list(data))

    @property
    def ndim(self) -> int:
        """
        Number of dimensions.
        """
        return len(self.scale)


class PathScale(Transform):
    """
    Scale transformation parametrized by a path.
    """

    type: Literal["scale"]
    path: str


class VectorTranslation(Transform):
    """
    Translation transformation parametrized by a vector of numbers.
    """

    type: Literal["translation"] = Field(..., description="Type")
    translation: list[float]

    @classmethod
    def build(cls, data: Iterable[float]) -> Self:
        """
        Create a VectorTranslation from an iterable of floats.
        """
        return cls(type="translation", translation=list(data))

    @property
    def ndim(self) -> int:
        """
        Number of dimensions.
        """
        return len(self.translation)


class PathTranslation(Transform):
    """
    Translation transformation parametrized by a path.
    """

    type: Literal["translation"]
    translation: str


ScaleTransform = VectorScale | PathScale
TranslationTransform = VectorTranslation | PathTranslation
VectorTransform = VectorScale | VectorTranslation


ValidTransform = tuple[ScaleTransform] | tuple[ScaleTransform, TranslationTransform]


def _ndim(transform: VectorTransform) -> int:
    """
    Get the dimensionality of a scale or translation transform.
    """
    return transform.ndim


def _build_transforms(
    scale: Sequence[float], translation: Sequence[float] | None
) -> tuple[VectorScale] | tuple[VectorScale, VectorTranslation]:
    """
    Create a `VectorScale` and optionally a `VectorTranslation` from a scale and a
    translation parameter.
    """
    vec_scale = VectorScale.build(scale)
    if translation is None:
        return (vec_scale,)
    else:
        vec_trans = VectorTranslation.build(translation)
        return vec_scale, vec_trans



from typing import TYPE_CHECKING

import zarr
from zarr.abc.store import Store

from ome_zarr_models.v05.hcs import HCS, HCSAttrs
from ome_zarr_models.v05.plate import Acquisition, Column, Plate, Row, WellInPlate
from tests.v05.conftest import json_to_zarr_group

if TYPE_CHECKING:
    from pydantic import JsonValue


def test_hcs(store: Store) -> None:
    zarr_group = json_to_zarr_group(json_fname="hcs_example.json", store=store)
    ome_group = HCS.from_zarr(zarr_group)
    assert ome_group.attributes.ome == HCSAttrs(
        plate=Plate(
            acquisitions=[
                Acquisition(
                    id=1,
                    name="Meas_01(2012-07-31_10-41-12)",
                    maximumfieldcount=2,
                    description=None,
                    starttime=1343731272000,
                    endtime=None,
                ),
                Acquisition(
                    id=2,
                    name="Meas_02(201207-31_11-56-41)",
                    maximumfieldcount=2,
                    description=None,
                    starttime=1343735801000,
                    endtime=None,
                ),
            ],
            columns=[Column(name="1"), Column(name="2"), Column(name="3")],
            field_count=4,
            name="test",
            rows=[Row(name="A"), Row(name="B")],
            wells=[
                WellInPlate(path="A/1", rowIndex=0, columnIndex=0),
                WellInPlate(path="A/2", rowIndex=0, columnIndex=1),
                WellInPlate(path="A/3", rowIndex=0, columnIndex=2),
                WellInPlate(path="B/1", rowIndex=1, columnIndex=0),
                WellInPlate(path="B/2", rowIndex=1, columnIndex=1),
                WellInPlate(path="B/3", rowIndex=1, columnIndex=2),
            ],
            version="0.5",
        ),
        version="0.5",
    )
    well_groups = list(ome_group.well_groups)
    assert len(well_groups) == 0


def test_non_existent_wells() -> None:
    """
    Make sure it's possible to create a HCS that has well paths that don't exist
    as Zarr groups.

    The relevant part of the OME-Zarr specification (https://ngff.openmicroscopy.org/0.5/index.html#plate-md)
    does not specify explicitly that the Zarr groups have to exist.
    """
    HCS(
        attributes={
            "ome": {
                "plate": {
                    "acquisitions": [{"id": 1}, {"id": 2}, {"id": 3}],
                    "columns": [{"name": "1"}],
                    "field_count": 10,
                    "rows": [{"name": "A"}],
                    "version": "0.4",
                    "wells": [{"columnIndex": 0, "path": "A/1", "rowIndex": 0}],
                },
                "version": "0.5",
            }
        }
    )


def test_non_existent_wells_from_zarr() -> None:
    """
    Same as above, but using from_zarr(...)
    """

    plate: dict[str, JsonValue] = {
        "columns": [{"name": "1"}],
        "rows": [{"name": "A"}],
        "wells": [{"path": "A/1", "rowIndex": 0, "columnIndex": 0}],
        "version": "0.5",
    }
    group = group = zarr.create_group(
        store={},
        zarr_format=3,
        attributes={"ome": {"plate": plate, "version": "0.5"}},
    )
    HCS.from_zarr(group)



# Need to import `annotations` for the pydantic_zarr TypeAlias strings to work
from __future__ import annotations

from typing import TYPE_CHECKING, Literal, TypeVar, overload

import zarr
import zarr.errors
from pydantic import StringConstraints
from pydantic_zarr.v2 import AnyArraySpec as AnyArraySpecv2
from pydantic_zarr.v2 import AnyGroupSpec as AnyGroupSpecv2
from pydantic_zarr.v2 import ArraySpec as ArraySpecv2
from pydantic_zarr.v2 import GroupSpec as GroupSpecv2
from pydantic_zarr.v3 import AnyArraySpec as AnyArraySpecv3
from pydantic_zarr.v3 import AnyGroupSpec as AnyGroupSpecv3
from pydantic_zarr.v3 import ArraySpec as ArraySpecv3
from pydantic_zarr.v3 import GroupSpec as GroupSpecv3

from ome_zarr_models.common.coordinate_transformations import VectorScale

if TYPE_CHECKING:
    from collections.abc import Sequence


__all__ = [
    "AlphaNumericConstraint",
    "RGBHexConstraint",
    "check_array_path",
    "unique_items_validator",
]

AlphaNumericConstraint = StringConstraints(pattern="^[a-zA-Z0-9]*$")
"""Require a string to only contain letters and numbers"""

RGBHexConstraint = StringConstraints(pattern=r"[0-9a-fA-F]{6}")
"""Require a string to be a valid RGB hex string"""

T = TypeVar("T")


def unique_items_validator(values: list[T]) -> list[T]:
    """
    Make sure a list contains unique items.

    Raises
    ------
    ValueError
        If duplicate values are found in *values*.
    """
    for ind, value in enumerate(values, start=1):
        if value in values[ind:]:
            raise ValueError(f"Duplicate values found in {values}.")
    return values


@overload
def check_array_path(
    group: zarr.Group,
    array_path: str,
    *,
    expected_zarr_version: Literal[2],
) -> AnyArraySpecv2: ...


@overload
def check_array_path(
    group: zarr.Group,
    array_path: str,
    *,
    expected_zarr_version: Literal[3],
) -> AnyArraySpecv3: ...


def check_array_path(
    group: zarr.Group,
    array_path: str,
    *,
    expected_zarr_version: Literal[2, 3],
) -> AnyArraySpecv2 | AnyArraySpecv3:
    """
    Check if an array exists at a given path in a group.

    Returns
    -------
    ArraySpec
        If the path exists, it's ArraySpec is returned.

    Raises
    ------
    ValueError
        If the array doesn't exist, or the array is not the expected Zarr version.
    """
    try:
        array = zarr.open_array(
            store=group.store_path,
            path=array_path,
            mode="r",
            zarr_format=expected_zarr_version,
        )
    except FileNotFoundError as e:
        msg = (
            f"Expected to find an array at {array_path}, but no array was found there."
        )
        raise ValueError(msg) from e
    except (zarr.errors.ContainsGroupError, zarr.errors.NodeTypeValidationError) as e:
        msg = (
            f"Expected to find an array at {array_path}, "
            "but a group was found there instead."
        )
        raise ValueError(msg) from e

    array_spec: AnyArraySpecv2 | AnyArraySpecv3
    if array.metadata.zarr_format == 2:
        if expected_zarr_version == 3:
            raise ValueError("Expected Zarr v3 array, but got v2 array")
        array_spec = ArraySpecv2.from_zarr(array)
    else:
        if expected_zarr_version == 2:
            raise ValueError("Expected Zarr v2 array, but got v3 array")
        array_spec = ArraySpecv3.from_zarr(array)

    return array_spec


@overload
def check_group_path(
    group: zarr.Group,
    group_path: str,
    *,
    expected_zarr_version: Literal[2],
) -> AnyGroupSpecv2: ...


@overload
def check_group_path(
    group: zarr.Group,
    group_path: str,
    *,
    expected_zarr_version: Literal[3],
) -> AnyGroupSpecv3: ...


def check_group_path(
    group: zarr.Group,
    group_path: str,
    *,
    expected_zarr_version: Literal[2, 3],
) -> AnyGroupSpecv2 | AnyGroupSpecv3:
    """
    Check if a group exists at a given path in a group.

    Returns
    -------
    GroupSpec
        If the path exists, it's GroupSpec is returned.

    Raises
    ------
    ValueError
        If the group doesn't exist, or the group is not the expected Zarr version.
    """
    try:
        group = zarr.open_group(
            store=group.store_path,
            path=group_path,
            mode="r",
            zarr_format=expected_zarr_version,
        )
    except FileNotFoundError as e:
        msg = f"Expected to find a group at {group_path}, but no group was found there."
        raise FileNotFoundError(msg) from e
    except zarr.errors.ContainsArrayError as e:
        msg = (
            f"Expected to find an group at {group_path}, "
            "but an array was found there instead."
        )
        raise zarr.errors.ContainsArrayError(msg) from e

    group_spec: AnyGroupSpecv2 | AnyGroupSpecv3
    if group.metadata.zarr_format == 2:
        if expected_zarr_version == 3:
            raise ValueError("Expected Zarr v3 array, but got v2 array")
        group_spec = GroupSpecv2.from_zarr(group, depth=0)
    else:
        if expected_zarr_version == 2:
            raise ValueError("Expected Zarr v2 array, but got v3 array")
        group_spec = GroupSpecv3.from_zarr(group, depth=0)

    return group_spec


def check_length(
    sequence: Sequence[T], *, valid_lengths: Sequence[int], variable_name: str
) -> None:
    """
    Check if the length of a sequence is valid.

    Raises
    ------
    ValueError
        If the sequence is not a valid length.
    """
    if len(sequence) not in valid_lengths:
        msg = (
            f"Length of {variable_name} ({len(sequence)}) not valid. "
            f"Allowed lengths are {valid_lengths}."
        )
        raise ValueError(msg)


@overload
def check_array_spec(spec: AnyGroupSpecv2, path: str) -> AnyArraySpecv2: ...


@overload
def check_array_spec(spec: AnyGroupSpecv3, path: str) -> AnyArraySpecv3: ...


def check_array_spec(
    spec: AnyGroupSpecv2 | AnyGroupSpecv3, path: str
) -> AnyArraySpecv2 | AnyArraySpecv3:
    """
    Check that a path within a group is an array.

    Raises
    ------
    ValueError
        If the node at *path* is not an array.
    """
    if spec.members is None:
        raise ValueError(f"members=None for {spec}")
    new_spec = spec.members[path]
    if not isinstance(new_spec, ArraySpecv2 | ArraySpecv3):
        raise ValueError(f"Node at path '{path}' is a group, expected an array")
    return new_spec


@overload
def check_group_spec(spec: AnyGroupSpecv2, path: str) -> AnyGroupSpecv2: ...


@overload
def check_group_spec(spec: AnyGroupSpecv3, path: str) -> AnyGroupSpecv3: ...


def check_group_spec(
    spec: AnyGroupSpecv2 | AnyGroupSpecv3, path: str
) -> AnyGroupSpecv2 | AnyGroupSpecv3:
    """
    Check that a path within a group is a group.

    Raises
    ------
    ValueError
        If the node at *path* is not a group.
    """
    if spec.members is None:
        raise ValueError("Specification has no members.")
    new_spec = spec.members[path]
    if not isinstance(new_spec, GroupSpecv2 | GroupSpecv3):
        raise ValueError(
            f"Node at path '{path}' is not a GroupSpec (got {type(new_spec)=})"
        )
    return new_spec


def check_ordered_scales(scales: list[VectorScale]) -> None:
    """
    Given a list of scales, make sure they are ordered from low to high.

    Raises
    ------
    ValueError
        If any items in one set of scales is smaller than any item in
        the preceding set of scales.
    """
    for i in range(len(scales) - 1):
        s1, s2 = scales[i].scale, scales[i + 1].scale
        is_ordered = all(s1[j] <= s2[j] for j in range(len(s1)))
        if not is_ordered:
            raise ValueError(
                f"Dataset {i} has a lower resolution (scales = {s1}) "
                f"than dataset {i + 1} (scales = {s2})."
            )



from collections.abc import Sequence
from typing import Self

# Import needed for pydantic type resolution
import pydantic_zarr  # noqa: F401
import zarr
import zarr.errors
from pydantic import Field, JsonValue, model_validator
from pydantic_zarr.v3 import AnyArraySpec, AnyGroupSpec, GroupSpec

from ome_zarr_models._utils import _from_zarr_v3
from ome_zarr_models.common.coordinate_transformations import _build_transforms
from ome_zarr_models.v05.axes import Axis
from ome_zarr_models.v05.base import BaseGroupv05, BaseOMEAttrs, BaseZarrAttrs
from ome_zarr_models.v05.labels import Labels
from ome_zarr_models.v05.multiscales import Dataset, Multiscale

__all__ = ["Image", "ImageAttrs"]


class ImageAttrs(BaseOMEAttrs):
    """
    Metadata for OME-Zarr image groups.
    """

    multiscales: list[Multiscale] = Field(
        ...,
        description="The multiscale datasets for this image",
        min_length=1,
    )

    def get_array_paths(self) -> list[str]:
        paths = []
        for multiscale in self.multiscales:
            for dataset in multiscale.datasets:
                paths.append(dataset.path)
        return paths

    def get_optional_group_paths(self) -> dict[str, type[Labels]]:  # type: ignore[override]
        return {"labels": Labels}


class Image(BaseGroupv05[ImageAttrs]):
    """
    An OME-Zarr image dataset.
    """

    @classmethod
    def from_zarr(cls, group: zarr.Group) -> Self:  # type: ignore[override]
        """
        Create an OME-Zarr image model from a `zarr.Group`.

        Parameters
        ----------
        group : zarr.Group
            A Zarr group that has valid OME-Zarr image metadata.
        """
        return _from_zarr_v3(group, cls, ImageAttrs)

    @classmethod
    def new(
        cls,
        *,
        array_specs: Sequence[AnyArraySpec],
        paths: Sequence[str],
        axes: Sequence[Axis],
        scales: Sequence[Sequence[float]],
        translations: Sequence[Sequence[float] | None],
        name: str | None = None,
        multiscale_type: str | None = None,
        metadata: JsonValue | None = None,
        global_scale: Sequence[float] | None = None,
        global_translation: Sequence[float] | None = None,
    ) -> "Image":
        """
        Create a new `Image` from a sequence of multiscale arrays
        and spatial metadata.

        Parameters
        ----------
        array_specs :
            A sequence of array specifications that collectively represent the same
            image at multiple levels of detail.
        paths :
            The paths to the arrays within the new Zarr group.
        axes :
            `Axis` objects describing the axes of the arrays.
        scales :
            For each array, a scale value for each axis of the array.
        translations :
            For each array, a translation value for each axis the array.
        name :
            A name for the multiscale collection.
        multiscale_type :
            Type of downscaling method used to generate the multiscale image pyramid.
            Optional.
        metadata :
            Arbitrary metadata to store in the multiscales group.
        global_scale :
            A global scale value for each axis of every array.
        global_translation :
            A global translation value for each axis of every array.

        Notes
        -----
        This class does not store or copy any array data. To save array data,
        first write this class to a Zarr store, and then write data to the Zarr
        arrays in that store.
        """
        if len(array_specs) != len(paths):
            raise ValueError(
                f"Length of arrays (got {len(array_specs)=}) must be the same as "
                f"length of paths (got {len(paths)=})"
            )
        members_flat = {
            "/" + key.lstrip("/"): arr
            for key, arr in zip(paths, array_specs, strict=True)
        }

        if global_scale is None and global_translation is None:
            global_transform = None
        elif global_scale is None:
            raise ValueError(
                "If global_translation is specified, "
                "global_scale must also be specified."
            )
        else:
            global_transform = _build_transforms(global_scale, global_translation)

        if len(scales) != len(paths):
            raise ValueError(
                f"Length of 'scales' ({len(scales)}) does not match "
                f"length of 'paths' {len(paths)}"
            )
        if len(translations) != len(paths):
            raise ValueError(
                f"Length of 'translations' ({len(translations)}) does not match "
                f"length of 'paths' ({len(paths)})"
            )
        multimeta = Multiscale(
            axes=tuple(axes),
            datasets=tuple(
                Dataset.build(path=path, scale=scale, translation=translation)
                for path, scale, translation in zip(
                    paths, scales, translations, strict=True
                )
            ),
            coordinateTransformations=global_transform,
            metadata=metadata,
            name=name,
            type=multiscale_type,
        )
        return Image(
            members=GroupSpec.from_flat(members_flat).members,
            attributes=BaseZarrAttrs(
                ome=ImageAttrs(
                    multiscales=[
                        multimeta,
                    ],
                    version="0.5",
                ),
            ),
        )

    @model_validator(mode="after")
    def _check_arrays_compatible(self) -> Self:
        """
        Check that all the arrays referenced by the `multiscales` metadata meet the
        following criteria:
            - they exist
            - they are not groups
            - they have dimensionality consistent with the number of axes defined in the
              metadata.
        """
        multimeta = self.ome_attributes.multiscales
        flat_self = self.to_flat()

        for multiscale in multimeta:
            multiscale_ndim = len(multiscale.axes)
            multiscale_dim_names = tuple(a.name for a in multiscale.axes)
            for dataset in multiscale.datasets:
                try:
                    maybe_arr: AnyArraySpec | AnyGroupSpec = flat_self[
                        "/" + dataset.path.lstrip("/")
                    ]
                except KeyError as e:
                    msg = (
                        f"The multiscale metadata references an array that does not "
                        f"exist in this group: {dataset.path}"
                    )
                    raise ValueError(msg) from e

                if isinstance(maybe_arr, GroupSpec):
                    msg = f"The node at {dataset.path} is a group, not an array."
                    raise ValueError(msg)

                arr_ndim = len(maybe_arr.shape)
                if arr_ndim != multiscale_ndim:
                    msg = (
                        f"The multiscale metadata has {multiscale_ndim} axes "
                        "which does not match the dimensionality of the array "
                        f"found in this group at path '{dataset.path}' ({arr_ndim}). "
                        "The number of axes must match the array dimensionality."
                    )

                    raise ValueError(msg)

                arr_dim_names = maybe_arr.dimension_names
                if arr_dim_names is None:
                    msg = (
                        f"The array in this group at  '{dataset.path}' has no "
                        "dimension_names metadata."
                    )
                    raise ValueError(msg)
                elif arr_dim_names != multiscale_dim_names:
                    msg = (
                        f"The multiscale metadata has {multiscale_dim_names} "
                        "axes names "
                        "which does not match the dimension names of the array "
                        f"found in this group at path '{dataset.path}' "
                        f"({arr_dim_names}). "
                    )

                    raise ValueError(msg)

        return self

    @model_validator(mode="after")
    def _check_label_multiscales(self) -> Self:
        """
        Check that the number of multiscale levels in any labels are the same as
        the base image.
        """
        if (labels := self.labels) is None:
            return self

        multimeta = self.ome_attributes.multiscales
        if len(multimeta) != 1:
            # If there's more than one multiscales, we can't check the labels
            # levels because we don't know which multiscales to check against
            # TODO: add a warning
            return self

        multiscales = multimeta[0]
        n_levels = len(multiscales.datasets)
        for path in labels.label_paths:
            image_label = labels.get_image_labels_group(path)
            if len(image_label.ome_attributes.multiscales) != 1:
                # If there's more than one multiscales, we can't check the labels
                # levels because we don't know which multiscales to check against
                # TODO: add a warning
                continue
            else:
                n_label_levels = len(image_label.ome_attributes.multiscales[0].datasets)
                if n_levels != n_label_levels:
                    raise RuntimeError(
                        f"Number of image label multiscale levels ({n_label_levels}) "
                        f"doesn't match number of image multiscale levels ({n_levels})."
                    )

        return self

    @property
    def labels(self) -> Labels | None:
        """
        Any labels datasets contained in this image group.

        Returns None if no labels are present.
        """
        if self.members is None or "labels" not in self.members:
            return None

        labels_group = self.members["labels"]
        if not isinstance(labels_group, GroupSpec):
            raise ValueError("Node at path 'labels' is not a group")

        return Labels(attributes=labels_group.attributes, members=labels_group.members)

    @property
    def datasets(self) -> tuple[tuple[Dataset, ...], ...]:
        """
        Get datasets stored in this image.

        The first index is for the multiscales.
        The second index is for the dataset inside that multiscales.
        """
        return tuple(
            tuple(dataset for dataset in multiscale.datasets)
            for multiscale in self.ome_attributes.multiscales
        )



from typing import TYPE_CHECKING, Any, Self

import numpy as np

# Import needed for pydantic type resolution
import pydantic_zarr  # noqa: F401
import zarr
import zarr.errors
from pydantic import Field, ValidationError, model_validator
from pydantic_zarr.v3 import AnyArraySpec, AnyGroupSpec, GroupSpec

from ome_zarr_models._v06.base import BaseGroupv06, BaseOMEAttrs
from ome_zarr_models.common.validation import check_array_spec, check_group_spec

if TYPE_CHECKING:
    from ome_zarr_models._v06.image_label import ImageLabel


__all__ = ["Labels", "LabelsAttrs"]


VALID_DTYPES: list[np.dtype[Any]] = [
    np.dtype(x)
    for x in [
        np.uint8,
        np.int8,
        np.uint16,
        np.int16,
        np.uint32,
        np.int32,
        np.uint64,
        np.int64,
    ]
]


def _check_valid_dtypes(labels: "Labels") -> "Labels":
    """
    Check that all multiscales levels of a labels image are valid Label data types.
    """
    from ome_zarr_models._v06.image_label import ImageLabel

    if labels.members is None:
        raise RuntimeError(f"{labels.members=}")

    for label_path in labels.attributes.ome.labels:
        if label_path not in labels.members:
            raise ValueError(f"Label path '{label_path}' not found in zarr group")
        label_spec = check_group_spec(labels, label_path)  # type: ignore[arg-type]
        try:
            image_spec = ImageLabel(
                attributes=label_spec.attributes, members=label_spec.members
            )
        except ValidationError as e:
            raise RuntimeError(
                f"Error validating multiscale image at path '{label_path}'. "
                "See above for more detailed error message."
            ) from e

        for multiscale in image_spec.attributes.ome.multiscales:
            for dataset in multiscale.datasets:
                arr_spec = check_array_spec(image_spec, dataset.path)  # type: ignore[arg-type]
                if (
                    not isinstance(arr_spec.data_type, str)
                    or np.dtype(arr_spec.data_type) not in VALID_DTYPES
                ):
                    msg = (
                        "Data type of labels at path "
                        f"'{label_path}/{dataset.path}' is not valid. "
                        f"Got {arr_spec.data_type}, should be one of "
                        f"{[str(x) for x in VALID_DTYPES]}."
                    )
                    raise ValueError(msg)

    return labels


class LabelsAttrs(BaseOMEAttrs):
    """
    Attributes for an OME-Zarr labels dataset.
    """

    labels: list[str] = Field(
        ..., description="List of paths to labels arrays within a labels dataset."
    )


class Labels(
    BaseGroupv06[LabelsAttrs],
):
    """
    An OME-Zarr labels dataset.
    """

    @classmethod
    def from_zarr(cls, group: zarr.Group) -> Self:  # type: ignore[override]
        """
        Create an OME-Zarr labels model from a `zarr.Group`.

        Parameters
        ----------
        group : zarr.Group
            A Zarr group that has valid OME-Zarr label metadata.
        """
        from ome_zarr_models._v06.image_label import ImageLabel

        attrs_dict = group.attrs.asdict()
        label_attrs = LabelsAttrs.model_validate(group.attrs.asdict()["ome"])

        # Extract Zarr Image paths from multiscale metadata
        members_tree_flat: dict[str, AnyGroupSpec | AnyArraySpec] = {}
        for label_path in label_attrs.labels:
            try:
                image_group = zarr.open_group(
                    store=group.store_path / label_path, mode="r", zarr_format=3
                )
            except zarr.errors.GroupNotFoundError as err:
                raise ValueError(
                    f"Label path '{label_path}' not found in zarr group"
                ) from err
            try:
                image_model = ImageLabel.from_zarr(image_group).to_flat()
            except Exception as err:
                msg = (
                    f"Error validating the label path '{label_path}' "
                    "as a OME-Zarr multiscales group."
                )
                raise RuntimeError(msg) from err
            for path in image_model:
                members_tree_flat["/" + label_path + path] = image_model[path]

        members_normalized: AnyGroupSpec = GroupSpec.from_flat(members_tree_flat)
        return cls(attributes=attrs_dict, members=members_normalized.members)

    _check_valid_dtypes = model_validator(mode="after")(_check_valid_dtypes)

    @property
    def label_paths(self) -> list[str]:
        """
        List of paths to image-label groups within this labels group.
        """
        return self.attributes.ome.labels

    def get_image_labels_group(self, path: str) -> "ImageLabel":
        """
        Get a image labels group at a given path.
        """
        from ome_zarr_models._v06.image_label import ImageLabel

        if self.members is None:
            raise RuntimeError(f"{self.members=}")
        spec = self.members[path]
        if not isinstance(spec, GroupSpec):
            raise RuntimeError(f"Node at {path} is not a group")

        return ImageLabel(attributes=spec.attributes, members=spec.members)



from typing import TYPE_CHECKING

import zarr

from ome_zarr_models.common.omero import Channel, Omero, Window
from ome_zarr_models.v04.axes import Axis
from ome_zarr_models.v04.coordinate_transformations import VectorScale
from ome_zarr_models.v04.hcs import HCS, HCSAttrs
from ome_zarr_models.v04.image import ImageAttrs
from ome_zarr_models.v04.multiscales import Dataset, Multiscale
from ome_zarr_models.v04.plate import Acquisition, Column, Plate, Row, WellInPlate
from ome_zarr_models.v04.well_types import WellImage, WellMeta
from tests.conftest import get_examples_path

if TYPE_CHECKING:
    from pydantic import JsonValue


def test_example_hcs() -> None:
    group = zarr.open_group(
        get_examples_path(version="0.4") / "hcs_example.ome.zarr",
        mode="r",
        zarr_format=2,
    )
    hcs: HCS = HCS.from_zarr(group)
    assert hcs.attributes == HCSAttrs(
        plate=Plate(
            acquisitions=[
                Acquisition(
                    id=0,
                    name="20200812-CardiomyocyteDifferentiation14-Cycle1",
                    maximumfieldcount=None,
                    description=None,
                    starttime=None,
                    endtime=None,
                )
            ],
            columns=[Column(name="03")],
            field_count=None,
            name=None,
            rows=[Row(name="B")],
            version="0.4",
            wells=[WellInPlate(path="B/03", rowIndex=0, columnIndex=0)],
        )
    )
    assert hcs.members is not None
    assert list(hcs.members.keys()) == ["B"]

    well_groups = list(hcs.well_groups)
    assert len(well_groups) == 1
    well_group = hcs.get_well_group(0)
    assert well_group.attributes.well == WellMeta(
        images=[WellImage(path="0", acquisition=None)], version="0.4"
    )

    images = list(well_group.images)
    assert len(images) == 1

    assert images[0].attributes == ImageAttrs(
        multiscales=[
            Multiscale(
                axes=[
                    Axis(name="c", type="channel", unit=None),
                    Axis(name="z", type="space", unit="micrometer"),
                    Axis(name="y", type="space", unit="micrometer"),
                    Axis(name="x", type="space", unit="micrometer"),
                ],
                datasets=[
                    Dataset(
                        path="0",
                        coordinateTransformations=(
                            VectorScale(type="scale", scale=[1.0, 1.0, 0.1625, 0.1625]),
                        ),
                    ),
                    Dataset(
                        path="1",
                        coordinateTransformations=(
                            VectorScale(type="scale", scale=[1.0, 1.0, 0.325, 0.325]),
                        ),
                    ),
                    Dataset(
                        path="2",
                        coordinateTransformations=(
                            VectorScale(type="scale", scale=[1.0, 1.0, 0.65, 0.65]),
                        ),
                    ),
                    Dataset(
                        path="3",
                        coordinateTransformations=(
                            VectorScale(type="scale", scale=[1.0, 1.0, 1.3, 1.3]),
                        ),
                    ),
                    Dataset(
                        path="4",
                        coordinateTransformations=(
                            VectorScale(type="scale", scale=[1.0, 1.0, 2.6, 2.6]),
                        ),
                    ),
                ],
                version="0.4",
                coordinateTransformations=None,
                metadata=None,
                name=None,
                type=None,
            )
        ],
        omero=Omero(
            channels=[
                Channel(
                    color="00FFFF",
                    window=Window(max=65535.0, min=0.0, start=110.0, end=800.0),
                    label="DAPI",
                    wavelength_id="A01_C01",
                )
            ],
            id=1,
            name="TBD",
            version="0.4",
        ),
    )


def test_non_existent_wells() -> None:
    """
    Make sure it's possible to create a HCS that has well paths that don't exist
    as Zarr groups.

    The relevant part of the OME-Zarr specification (https://ngff.openmicroscopy.org/0.4/index.html#plate-md)
    does not specify explicitly that the Zarr groups have to exist.
    """
    HCS(
        attributes={
            "plate": {
                "acquisitions": [{"id": 1}, {"id": 2}, {"id": 3}],
                "columns": [{"name": "1"}],
                "field_count": 10,
                "rows": [{"name": "A"}],
                "version": "0.4",
                "wells": [{"columnIndex": 0, "path": "A/1", "rowIndex": 0}],
            }
        }
    )


def test_non_existent_wells_from_zarr() -> None:
    """
    Same as above, but using from_zarr(...)
    """
    plate: dict[str, JsonValue] = {
        "columns": [{"name": "1"}],
        "rows": [{"name": "A"}],
        "wells": [{"path": "A/1", "rowIndex": 0, "columnIndex": 0}],
        "version": "0.4",
    }
    group = group = zarr.create_group(
        store={},
        zarr_format=2,
        attributes={"plate": plate, "version": "0.4"},
    )
    HCS.from_zarr(group)



from __future__ import annotations

from collections import Counter
from typing import TYPE_CHECKING, Any, Self

from pydantic import (
    BaseModel,
    Field,
    JsonValue,
    SerializerFunctionWrapHandler,
    field_validator,
    model_serializer,
    model_validator,
)

from ome_zarr_models._v06.axes import Axes
from ome_zarr_models.base import BaseAttrs
from ome_zarr_models.common.coordinate_transformations import (
    Transform,
    ValidTransform,
    VectorScale,
    VectorTransform,
    _build_transforms,
    _ndim,
)
from ome_zarr_models.common.validation import (
    check_length,
    check_ordered_scales,
    unique_items_validator,
)

if TYPE_CHECKING:
    from collections.abc import Sequence


__all__ = ["Dataset", "Multiscale"]


VALID_NDIM = (2, 3, 4, 5)


class Multiscale(BaseAttrs):
    """
    An element of multiscales metadata.
    """

    axes: Axes
    datasets: tuple[Dataset, ...] = Field(..., min_length=1)
    coordinateTransformations: ValidTransform | None = None
    metadata: JsonValue = None
    name: JsonValue | None = None
    type: JsonValue = None

    @model_serializer(mode="wrap")
    def _serialize(
        self,
        serializer: SerializerFunctionWrapHandler,
    ) -> dict[str, Any]:
        d: dict[str, Any] = serializer(self)
        if self.coordinateTransformations is None:
            d.pop("coordinateTransformations", None)

        return d

    @property
    def ndim(self) -> int:
        """
        Dimensionality of the data described by this metadata.

        Determined by the length of the axes attribute.
        """
        return len(self.axes)

    @model_validator(mode="after")
    def _ensure_axes_top_transforms(data: Self) -> Self:
        """
        Ensure that the length of the axes matches the dimensionality of the transforms
        defined in the top-level coordinateTransformations, if present.
        """
        if data.coordinateTransformations is not None:
            for tx in data.coordinateTransformations:
                if hasattr(tx, "ndim") and data.ndim != tx.ndim:
                    msg = (
                        f"The length of axes does not match the dimensionality of "
                        f"the {tx.type} transform in coordinateTransformations. "
                        f"Got {data.ndim} axes, but the {tx.type} transform has "
                        f"dimensionality {tx.ndim}"
                    )
                    raise ValueError(msg)
        return data

    @model_validator(mode="after")
    def _ensure_axes_dataset_transforms(data: Self) -> Self:
        """
        Ensure that the length of the axes matches the dimensionality of the transforms
        """
        self_ndim = len(data.axes)
        for ds_idx, ds in enumerate(data.datasets):
            for tx in ds.coordinateTransformations:
                if hasattr(tx, "ndim") and self_ndim != tx.ndim:
                    msg = (
                        f"The length of axes does not match the dimensionality of "
                        f"the {tx.type} transform in "
                        f"datasets[{ds_idx}].coordinateTransformations. "
                        f"Got {self_ndim} axes, but the {tx.type} transform has "
                        f"dimensionality {tx.ndim}"
                    )
                    raise ValueError(msg)
        return data

    @field_validator("datasets", mode="after")
    @classmethod
    def _ensure_ordered_scales(cls, datasets: list[Dataset]) -> list[Dataset]:
        """
        Make sure datasets are ordered from highest resolution to smallest.
        """
        scale_transforms = [d.coordinateTransformations[0] for d in datasets]
        # Only handle scales given in metadata, not in files
        scale_vector_transforms = [
            t for t in scale_transforms if isinstance(t, VectorScale)
        ]
        check_ordered_scales(scale_vector_transforms)
        return datasets

    @field_validator("axes", mode="after")
    @classmethod
    def _ensure_axis_length(cls, axes: Axes) -> Axes:
        """
        Ensures that there are between 2 and 5 axes (inclusive)
        """
        check_length(axes, valid_lengths=VALID_NDIM, variable_name="axes")
        return axes

    @field_validator("axes", mode="after")
    @classmethod
    def _ensure_axis_types(cls, axes: Axes) -> Axes:
        """
        Ensures that the following conditions are true:

        - there are only 2 or 3 axes with type `space`
        - the axes with type `space` are last in the list of axes
        - there is only 1 axis with type `time`
        - there is only 1 axis with type `channel`
        - there is only 1 axis with a type that is not `space`, `time`, or `channel`
        """
        check_length(
            [ax for ax in axes if ax.type == "space"],
            valid_lengths=[2, 3],
            variable_name="space axes",
        )
        check_length(
            [ax for ax in axes if ax.type == "time"],
            valid_lengths=[0, 1],
            variable_name="time axes",
        )
        check_length(
            [ax for ax in axes if ax.type == "channel"],
            valid_lengths=[0, 1],
            variable_name="channel axes",
        )
        check_length(
            [ax for ax in axes if ax.type not in ["space", "time", "channel"]],
            valid_lengths=[0, 1],
            variable_name="custom axes",
        )

        axis_types = [ax.type for ax in axes]
        type_census = Counter(axis_types)
        num_spaces = type_census["space"]
        if not all(a == "space" for a in axis_types[-num_spaces:]):
            msg = (
                f"All space axes must be at the end of the axes list. "
                f"Got axes with order: {axis_types}."
            )
            raise ValueError(msg)

        num_times = type_census["time"]
        if num_times == 1 and axis_types[0] != "time":
            msg = "Time axis must be at the beginning of axis list."
            raise ValueError(msg)

        return axes

    @field_validator("axes", mode="after")
    @classmethod
    def _ensure_unique_axis_names(cls, axes: Axes) -> Axes:
        """
        Ensures that the names of the axes are unique.
        """
        try:
            unique_items_validator(axis_names := [a.name for a in axes])
        except ValueError:
            raise ValueError(f"Axis names must be unique. Got {axis_names}") from None
        return axes


class Dataset(BaseAttrs):
    """
    An element of Multiscale.datasets.
    """

    # TODO: validate that path resolves to an actual zarr array
    # TODO: can we validate that the paths must be ordered from highest resolution to
    # smallest using scale metadata?
    path: str
    coordinateTransformations: ValidTransform

    @classmethod
    def build(
        cls, *, path: str, scale: Sequence[float], translation: Sequence[float] | None
    ) -> Self:
        """
        Construct a `Dataset` from a path, a scale, and a translation.
        """
        return cls(
            path=path,
            coordinateTransformations=_build_transforms(
                scale=scale, translation=translation
            ),
        )

    @field_validator("coordinateTransformations", mode="before")
    def _ensure_scale_translation(
        transforms_obj: object,
    ) -> object:
        """
        Ensures that
        - there are only 1 or 2 transforms.
        - the first element is a scale transformation
        - the second element, if present, is a translation transform
        """
        # This is used as a before validator - to help use, we use pydantic to first
        # cast the input (which can in general anything) into a set of transformations.
        # Then we check the transformations are valid.
        #
        # This is a bit convoluted, but we do it because the default pydantic error
        # messages are a mess otherwise

        class Transforms(BaseModel):
            transforms: list[Transform]

        transforms = Transforms(transforms=transforms_obj).transforms
        check_length(transforms, valid_lengths=[1, 2], variable_name="transforms")

        maybe_scale = transforms[0]
        if maybe_scale.type != "scale":
            msg = (
                "The first element of `coordinateTransformations` must be a scale "
                f"transform. Got {maybe_scale} instead."
            )
            raise ValueError(msg)
        if len(transforms) == 2:
            maybe_trans = transforms[1]
            if (maybe_trans.type) != "translation":
                msg = (
                    "The second element of `coordinateTransformations` must be a "
                    f"translation transform. Got {maybe_trans} instead."
                )
                raise ValueError(msg)

        return transforms_obj

    @field_validator("coordinateTransformations", mode="after")
    @classmethod
    def _ensure_transform_dimensionality(
        cls,
        transforms: ValidTransform,
    ) -> ValidTransform:
        """
        Ensures that the elements in the input sequence define transformations with
        identical dimensionality. If any of the transforms are defined with a path
        instead of concrete values, then no validation will be performed and the
        transforms will be returned as-is.
        """
        vector_transforms = filter(lambda v: isinstance(v, VectorTransform), transforms)
        ndims = tuple(map(_ndim, vector_transforms))  # type: ignore[arg-type]
        ndims_set = set(ndims)
        if len(ndims_set) > 1:
            msg = (
                "The transforms have inconsistent dimensionality. "
                f"Got transforms with dimensionality = {ndims}."
            )
            raise ValueError(msg)
        return transforms



from ome_zarr_models.v04._bioformats2raw import BioFormats2RawAttrs
from ome_zarr_models.v04.plate import (
    Acquisition,
    Column,
    Plate,
    Row,
    WellInPlate,
)
from tests.v04.conftest import read_in_json


def test_bioformats2raw_example_json() -> None:
    model = read_in_json(
        json_fname="bioformats2raw_example.json", model_cls=BioFormats2RawAttrs
    )

    assert model == BioFormats2RawAttrs(
        bioformats2raw_layout=3,
        plate=Plate(
            acquisitions=[
                Acquisition(id=0, maximumfieldcount=None, name=None, description=None)
            ],
            columns=[Column(name="1")],
            field_count=1,
            name="Plate Name 0",
            rows=[Row(name="A")],
            version="0.4",
            wells=[WellInPlate(path="A/1", rowIndex=0, columnIndex=0)],
        ),
        series=None,
    )



import re

import pytest
from pydantic import ValidationError

from ome_zarr_models.v04.plate import Acquisition, Column, Plate, Row, WellInPlate
from tests.v04.conftest import read_in_json


def test_example_plate_json() -> None:
    plate = read_in_json(json_fname="plate_example_1.json", model_cls=Plate)
    assert plate == Plate(
        acquisitions=[
            Acquisition(
                id=1,
                name="Meas_01(2012-07-31_10-41-12)",
                maximumfieldcount=2,
                description=None,
                starttime=1343731272000,
                endtime=None,
            ),
            Acquisition(
                id=2,
                name="Meas_02(201207-31_11-56-41)",
                maximumfieldcount=2,
                description=None,
                starttime=1343735801000,
                endtime=None,
            ),
        ],
        columns=[Column(name="1"), Column(name="2"), Column(name="3")],
        field_count=4,
        name="test",
        rows=[Row(name="A"), Row(name="B")],
        version="0.4",
        wells=[
            WellInPlate(path="A/1", rowIndex=0, columnIndex=0),
            WellInPlate(path="A/2", rowIndex=0, columnIndex=1),
            WellInPlate(path="A/3", rowIndex=0, columnIndex=2),
            WellInPlate(path="B/1", rowIndex=1, columnIndex=0),
            WellInPlate(path="B/2", rowIndex=1, columnIndex=1),
            WellInPlate(path="B/3", rowIndex=1, columnIndex=2),
        ],
    )


def test_example_plate_json_2() -> None:
    plate = read_in_json(json_fname="plate_example_2.json", model_cls=Plate)
    assert plate == Plate(
        acquisitions=[
            Acquisition(
                id=1,
                name="single acquisition",
                maximumfieldcount=1,
                description=None,
                starttime=1343731272000,
                endtime=None,
            )
        ],
        columns=[
            Column(name="1"),
            Column(name="2"),
            Column(name="3"),
            Column(name="4"),
            Column(name="5"),
            Column(name="6"),
            Column(name="7"),
            Column(name="8"),
            Column(name="9"),
            Column(name="10"),
            Column(name="11"),
            Column(name="12"),
        ],
        field_count=1,
        name="sparse test",
        rows=[
            Row(name="A"),
            Row(name="B"),
            Row(name="C"),
            Row(name="D"),
            Row(name="E"),
            Row(name="F"),
            Row(name="G"),
            Row(name="H"),
        ],
        version="0.4",
        wells=[
            WellInPlate(path="C/5", rowIndex=2, columnIndex=4),
            WellInPlate(path="D/7", rowIndex=3, columnIndex=6),
        ],
    )


def test_unique_column_names() -> None:
    with pytest.raises(ValidationError, match="Duplicate values found in"):
        Plate(
            columns=[Column(name="col1"), Column(name="col1")],
            rows=[Row(name="row1")],
            version="0.4",
            wells=[WellInPlate(path="path1", rowIndex=1, columnIndex=1)],
        )


def test_unique_row_names() -> None:
    with pytest.raises(ValidationError, match="Duplicate values found in"):
        Plate(
            columns=[Column(name="col1")],
            rows=[Row(name="row1"), Row(name="row1")],
            version="0.4",
            wells=[WellInPlate(path="path1", rowIndex=1, columnIndex=1)],
        )


@pytest.mark.parametrize("cls", [Row, Column])
def test_alphanumeric_column_names(cls: type[Row | Column]) -> None:
    with pytest.raises(ValidationError, match="String should match pattern "):
        cls(name="col-1")


@pytest.mark.parametrize(
    ("well_path", "msg"),
    [
        ("path1", "well path 'path1' does not contain a single '/'"),
        ("row1/col1/", "well path 'row1/col1/' does not contain a single '/'"),
        (
            "row1/col2",
            "column 'col2' in well path 'row1/col2' is not in list of columns",
        ),
        (
            "row2/col1",
            "row 'row2' in well path 'row2/col1' is not in list of rows",
        ),
    ],
)
def test_well_paths(well_path: str, msg: str) -> None:
    # No separator
    with pytest.raises(ValidationError, match=re.escape(msg)):
        Plate(
            columns=[Column(name="col1")],
            rows=[Row(name="row1")],
            version="0.4",
            wells=[WellInPlate(path=well_path, rowIndex=1, columnIndex=1)],
        )



from zarr.abc.store import Store

from ome_zarr_models.v04.image import Image
from ome_zarr_models.v04.labels import LabelsAttrs
from tests.v04.conftest import json_to_zarr_group


def test_image_with_labels(store: Store) -> None:
    zarr_group = json_to_zarr_group(json_fname="multiscales_example.json", store=store)
    zarr_group.create_group("labels")
    zarr_group["labels"].create_array("labels0", shape=(1, 1), dtype="uint8")  # type: ignore[union-attr]
    zarr_group["labels"].attrs.put({"labels": ["labels0"]})

    zarr_group.create_array("0", shape=(1, 1, 1, 1), dtype="uint8")
    zarr_group.create_array("1", shape=(1, 1, 1, 1), dtype="uint8")

    ome_group = Image.from_zarr(zarr_group)
    labels_group = ome_group.labels
    assert labels_group is not None
    assert labels_group.attributes == LabelsAttrs(labels=["labels0"])



"""
For reference, see the [plate section of the OME-Zarr specification](https://ngff.openmicroscopy.org/0.5/index.html#plate-md).
"""

import warnings
from typing import Self

from pydantic import Field, model_validator

from ome_zarr_models.common.plate import (
    Acquisition,
    Column,
    PlateBase,
    Row,
    WellInPlate,
)
from ome_zarr_models.exceptions import ValidationWarning

__all__ = [
    "Acquisition",
    "Column",
    "PlateBase",
    "Row",
    "WellInPlate",
]


class Plate(PlateBase):
    """
    A single plate.
    """

    version: str = Field(
        default="0.5", description="Version of the plate specification"
    )

    @model_validator(mode="after")
    def check_version_given(self) -> Self:
        if "version" not in self.model_fields_set:
            warnings.warn(
                "'version' field not specified in plate metadata, "
                "setting version='0.5'",
                ValidationWarning,
                stacklevel=2,
            )
        return self



"""
For reference, see the [well section of the OME-Zarr specification](https://ngff.openmicroscopy.org/0.4/#well-md).
"""

from collections.abc import Generator
from typing import Self

import zarr
from pydantic_zarr.v2 import AnyGroupSpec

from ome_zarr_models._utils import _from_zarr_v2
from ome_zarr_models.base import BaseAttrsv2
from ome_zarr_models.v04.base import BaseGroupv04
from ome_zarr_models.v04.image import Image
from ome_zarr_models.v04.well_types import WellMeta

__all__ = ["Well", "WellAttrs"]


class WellAttrs(BaseAttrsv2):
    """
    Attributes for a well group.
    """

    well: WellMeta

    def get_optional_group_paths(self) -> dict[str, type[AnyGroupSpec]]:
        return {im.path: Image for im in self.well.images}


class Well(BaseGroupv04[WellAttrs]):
    """
    An OME-Zarr well group.
    """

    @classmethod
    def from_zarr(cls, group: zarr.Group) -> Self:  # type: ignore[override]
        """
        Create an OME-Zarr well model from a `zarr.Group`.

        Parameters
        ----------
        group : zarr.Group
            A Zarr group that has valid OME-Zarr well metadata.
        """
        return _from_zarr_v2(group, cls, WellAttrs)

    def get_image(self, i: int) -> Image:
        """
        Get a single image from this well.
        """
        image = self.attributes.well.images[i]
        image_path = image.path
        image_path_parts = image_path.split("/")
        group: AnyGroupSpec = self
        for part in image_path_parts:
            if group.members is None:
                raise RuntimeError(f"{group.members=}")
            group = group.members[part]

        return Image(attributes=group.attributes, members=group.members)

    @property
    def n_images(self) -> int:
        """
        Number of images.
        """
        return len(self.attributes.well.images)

    @property
    def images(self) -> Generator[Image, None, None]:
        """
        Generator for all images in this well.
        """
        for i in range(self.n_images):
            yield self.get_image(i)



import re

import numpy as np
import pytest
from zarr.abc.store import Store

from ome_zarr_models._v06.labels import Labels, LabelsAttrs
from tests._v06.conftest import json_to_zarr_group


def test_labels(store: Store) -> None:
    zarr_group = json_to_zarr_group(json_fname="labels_example.json", store=store)
    image_group = zarr_group.create_group("cell_space_segmentation")
    image_group.attrs.put(
        json_to_zarr_group(
            json_fname="labels_image_example.json", store=store
        ).attrs.asdict()
    )
    image_group.create_array(
        "0",
        shape=(1, 1, 1, 1, 1),
        dtype=np.uint64,
        dimension_names=["t", "c", "z", "y", "x"],
    )

    ome_group = Labels.from_zarr(zarr_group)
    assert ome_group.attributes.ome == LabelsAttrs(
        labels=["cell_space_segmentation"], version="0.6"
    )


def test_labels_no_images(store: Store) -> None:
    zarr_group = json_to_zarr_group(json_fname="labels_example.json", store=store)
    with pytest.raises(
        ValueError,
        match="Label path 'cell_space_segmentation' not found in zarr group",
    ):
        Labels.from_zarr(zarr_group)


def test_labels_invalid_dtype(store: Store) -> None:
    """
    Check that an invalid data type raises an error.
    """
    zarr_group = json_to_zarr_group(json_fname="labels_example.json", store=store)
    image_group = zarr_group.create_group("cell_space_segmentation")
    image_group.attrs.put(
        json_to_zarr_group(
            json_fname="labels_image_example.json", store=store
        ).attrs.asdict()
    )
    image_group.create_array(
        "0",
        shape=(1, 1, 1, 1, 1),
        dtype=np.float32,
        dimension_names=["t", "c", "z", "y", "x"],
    )

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Data type of labels at path 'cell_space_segmentation/0' is not valid. "
            "Got float32, should be one of "
            "['uint8', 'int8', 'uint16', 'int16', 'uint32', 'int32', 'uint64', 'int64']"
        ),
    ):
        Labels.from_zarr(zarr_group)



"""
Custom exceptions and warnings used by `ome-zarr-models`
"""


class ValidationWarning(UserWarning):
    """
    Warning emitted for OME-Zarr data that can be interpreted
    by `ome-zarr-models`, but is not strictly compliant with the specification.
    """



import re

import numpy as np
import pytest
from pydantic_zarr.v2 import ArraySpec
from zarr.abc.store import Store

from ome_zarr_models.common.coordinate_transformations import VectorTranslation
from ome_zarr_models.v04.axes import Axis
from ome_zarr_models.v04.coordinate_transformations import VectorScale
from ome_zarr_models.v04.image import Image, ImageAttrs
from ome_zarr_models.v04.multiscales import Dataset, Multiscale
from tests.v04.conftest import json_to_zarr_group


def test_image(store: Store) -> None:
    zarr_group = json_to_zarr_group(json_fname="multiscales_example.json", store=store)
    zarr_group.create_array(
        "0",
        shape=(1, 1, 1, 1),
        dtype="uint8",
    )
    zarr_group.create_array("1", shape=(1, 1, 1, 1), dtype="uint8")

    ome_group = Image.from_zarr(zarr_group)
    assert ome_group.attributes == ImageAttrs(
        multiscales=[
            Multiscale(
                axes=[
                    Axis(name="c", type="channel", unit=None),
                    Axis(name="z", type="space", unit="micrometer"),
                    Axis(name="y", type="space", unit="micrometer"),
                    Axis(name="x", type="space", unit="micrometer"),
                ],
                datasets=(
                    Dataset(
                        path="0",
                        coordinateTransformations=(
                            VectorScale(
                                type="scale",
                                scale=[
                                    1.0,
                                    0.5002025531914894,
                                    0.3603981534640209,
                                    0.3603981534640209,
                                ],
                            ),
                        ),
                    ),
                    Dataset(
                        path="1",
                        coordinateTransformations=(
                            VectorScale(
                                type="scale",
                                scale=[
                                    1.0,
                                    0.5002025531914894,
                                    0.7207963069280418,
                                    0.7207963069280418,
                                ],
                            ),
                        ),
                    ),
                ),
                version="0.4",
                coordinateTransformations=None,
                metadata=None,
                name=None,
                type=None,
            )
        ],
        omero=None,
        _creator={"name": "omero-zarr", "version": "0.4.0"},
    )


def test_new_image() -> None:
    new_image = Image.new(
        array_specs=[
            ArraySpec(shape=(5, 5), chunks=(2, 2), dtype=np.uint8, attributes={}),
            ArraySpec(shape=(3, 3), chunks=(2, 2), dtype=np.uint8, attributes={}),
        ],
        paths=["scale0", "scale1"],
        axes=[
            Axis(name="x", type="space", unit="km"),
            Axis(name="y", type="space", unit="km"),
        ],
        scales=[(4, 4), (8, 8)],
        translations=[(2, 2), (4, 4)],
        name="new_image_test",
        multiscale_type="local mean",
        metadata={"key": "val"},
        global_scale=(-1, 1),
        global_translation=(10, 10),
    )
    assert new_image == Image(
        attributes=ImageAttrs(
            multiscales=[
                Multiscale(
                    axes=(
                        Axis(name="x", type="space", unit="km"),
                        Axis(name="y", type="space", unit="km"),
                    ),
                    datasets=(
                        Dataset(
                            path="scale0",
                            coordinateTransformations=(
                                VectorScale(type="scale", scale=[4.0, 4.0]),
                                VectorTranslation(
                                    type="translation", translation=[2.0, 2.0]
                                ),
                            ),
                        ),
                        Dataset(
                            path="scale1",
                            coordinateTransformations=(
                                VectorScale(type="scale", scale=[8.0, 8.0]),
                                VectorTranslation(
                                    type="translation", translation=[4.0, 4.0]
                                ),
                            ),
                        ),
                    ),
                    coordinateTransformations=(
                        VectorScale(type="scale", scale=[-1.0, 1.0]),
                        VectorTranslation(type="translation", translation=[10.0, 10.0]),
                    ),
                    metadata={"key": "val"},
                    name="new_image_test",
                    type="local mean",
                    version="0.4",
                )
            ],
            omero=None,
        ),
        members={
            "scale0": ArraySpec(
                attributes={},
                shape=(5, 5),
                chunks=(2, 2),
                dtype="|u1",
                fill_value=0,
                order="C",
                filters=None,
                dimension_separator="/",
                compressor=None,
            ),
            "scale1": ArraySpec(
                attributes={},
                shape=(3, 3),
                chunks=(2, 2),
                dtype="|u1",
                fill_value=0,
                order="C",
                filters=None,
                dimension_separator="/",
                compressor=None,
            ),
        },
    )


@pytest.fixture
def example_image() -> Image:
    return Image.new(
        array_specs=[
            ArraySpec(shape=(5, 5), chunks=(2, 2), dtype=np.uint8, attributes={}),
            ArraySpec(shape=(3, 3), chunks=(2, 2), dtype=np.uint8, attributes={}),
        ],
        paths=["scale0", "scale1"],
        axes=[
            Axis(name="x", type="space", unit="km"),
            Axis(name="y", type="space", unit="km"),
        ],
        scales=[(4, 4), (8, 8)],
        translations=[(2, 2), (4, 4)],
        name="new_image_test",
        multiscale_type="local mean",
        metadata={"key": "val"},
        global_scale=(-1, 1),
        global_translation=(10, 10),
    )


def test_datasets(example_image: Image) -> None:
    assert example_image.datasets == (
        (
            Dataset(
                path="scale0",
                coordinateTransformations=(
                    VectorScale(type="scale", scale=[4.0, 4.0]),
                    VectorTranslation(type="translation", translation=[2.0, 2.0]),
                ),
            ),
            Dataset(
                path="scale1",
                coordinateTransformations=(
                    VectorScale(type="scale", scale=[8.0, 8.0]),
                    VectorTranslation(type="translation", translation=[4.0, 4.0]),
                ),
            ),
        ),
    )


def test_new_image_wrong_transforms() -> None:
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Length of 'translations' (1) does not match length of 'paths' (2)"
        ),
    ):
        Image.new(
            array_specs=[
                ArraySpec(shape=(5, 5), chunks=(2, 2), dtype=np.uint8, attributes={}),
                ArraySpec(shape=(3, 3), chunks=(2, 2), dtype=np.uint8, attributes={}),
            ],
            paths=["scale0", "scale1"],
            axes=[
                Axis(name="x", type="space", unit="km"),
                Axis(name="y", type="space", unit="km"),
            ],
            scales=[(4, 4), (8, 8)],
            translations=[(2, 2)],
            name="new_image_test",
            multiscale_type="local mean",
            metadata={"key": "val"},
            global_scale=(-1, 1),
            global_translation=(10, 10),
        )


def test_global_transform(example_image: Image) -> None:
    model_dict = example_image.model_dump()
    assert "coordinateTransformations" in model_dict["attributes"]["multiscales"][0]


def test_no_global_transform() -> None:
    new_image = Image.new(
        array_specs=[
            ArraySpec(shape=(5, 5), chunks=(2, 2), dtype=np.uint8, attributes={}),
            ArraySpec(shape=(3, 3), chunks=(2, 2), dtype=np.uint8, attributes={}),
        ],
        paths=["scale0", "scale1"],
        axes=[
            Axis(name="x", type="space", unit="km"),
            Axis(name="y", type="space", unit="km"),
        ],
        scales=[(4, 4), (8, 8)],
        translations=[(2, 2), (4, 4)],
    )
    model_dict = new_image.model_dump()
    assert "coordinateTransformations" not in model_dict["attributes"]["multiscales"][0]

    new_image.model_dump(exclude_none=True)



from importlib.metadata import PackageNotFoundError, version
from typing import TYPE_CHECKING, Any, Literal

import zarr
import zarr.storage

import ome_zarr_models.v04.hcs
import ome_zarr_models.v04.image
import ome_zarr_models.v04.image_label
import ome_zarr_models.v04.labels
import ome_zarr_models.v04.well
import ome_zarr_models.v05.hcs
import ome_zarr_models.v05.image
import ome_zarr_models.v05.image_label
import ome_zarr_models.v05.labels
import ome_zarr_models.v05.well
from ome_zarr_models.base import BaseGroup
from ome_zarr_models.v04.base import BaseGroupv04
from ome_zarr_models.v05.base import BaseGroupv05

try:
    __version__ = version("ome_zarr_models")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "uninstalled"

if TYPE_CHECKING:
    from collections.abc import Sequence

_V04_groups: list[type[BaseGroupv04[Any]]] = [
    ome_zarr_models.v04.hcs.HCS,
    # Important that ImageLabel is higher than Image
    # otherwise Image will happily parse an ImageLabel
    # dataset without parsing the image-label bit of
    # metadata
    ome_zarr_models.v04.image_label.ImageLabel,
    ome_zarr_models.v04.image.Image,
    ome_zarr_models.v04.labels.Labels,
    ome_zarr_models.v04.well.Well,
]

_V05_groups: list[type[BaseGroupv05[Any]]] = [
    ome_zarr_models.v05.hcs.HCS,
    # ImageLabel does not appear here, as it is impossible to tell the
    # difference between an ImageLabel and Image group from the metadata
    #
    # Instead some custom logic is used to try and construct an
    # ImageLabel object in open_ome_zarr() below.
    #
    # See https://github.com/ome/ngff/issues/339 for more information
    # and discussion on this change from OME-Zarr 0.4
    ome_zarr_models.v05.image.Image,
    ome_zarr_models.v05.labels.Labels,
    ome_zarr_models.v05.well.Well,
]

_ome_zarr_zarr_map: dict[str, Literal[2, 3]] = {
    "0.4": 2,
    "0.5": 3,
}


_AnyGroup = type[BaseGroupv05[Any] | BaseGroupv04[Any]]


def open_ome_zarr(
    group: zarr.Group | zarr.storage.StoreLike,
    *,
    version: Literal["0.4", "0.5"] | None = None,
) -> BaseGroup:
    """
    Create an ome-zarr-models object from an existing OME-Zarr group.

    This function will 'guess' which type of OME-Zarr data exists by
    trying to validate each group metadata definition against your data.
    If validation is successful, that data class is returned without
    trying any more.

    It tries more recent versions of OME-Zarr first.

    Parameters
    ----------
    group : zarr.Group, zarr.storage.StoreLike
        Zarr group containing OME-Zarr data. Alternatively any object that can be
        parsed by [zarr.open_group][].
    version : Literal['0.4', '0.5'], optional
        If you know which version of OME-Zarr your data is, you can
        specify it here. If not specified, all versions will be tried.
        The default is None, which means all versions will be tried.

    Raises
    ------
    RuntimeError
        If the passed group cannot be validated with any of the OME-Zarr group models.

    Warnings
    --------
    This will try and load your data with every version of every OME-Zarr group type,
    until a match is found. If data access is slow (e.g., in a remote store), this may
    take a long time. It will be quicker to directly use the OME-Zarr group class if you
    know which version and group you expect.
    """
    if not isinstance(group, zarr.Group):
        zarr_format = _ome_zarr_zarr_map.get(version, None)  # type: ignore[arg-type]
        group = zarr.open_group(group, zarr_format=zarr_format, mode="r")

    # because 'from_zarr' isn't defined on a shared super-class, list all variants here
    groups: Sequence[_AnyGroup]
    match version:
        case None:
            groups = [*_V05_groups, *_V04_groups]
        case "0.4":
            groups = _V04_groups
        case "0.5":
            groups = _V05_groups
        case _:
            _versions = ("0.4", "0.5")  # type: ignore[unreachable]
            raise ValueError(
                f"Unsupported version '{version}', must be one of {_versions}, or None"
            )

    errors: list[tuple[_AnyGroup, Exception]] = []
    grp = None
    for group_cls in groups:
        try:
            grp = group_cls.from_zarr(group)
            break
        except Exception as e:
            errors.append((group_cls, e))

    # See if we have ImageLabel instead of an Image
    if (
        isinstance(grp, ome_zarr_models.v05.image.Image)
        and "image-label" in grp.ome_attributes.model_dump()
    ):
        try:
            return ome_zarr_models.v05.image_label.ImageLabel(
                attributes=grp.attributes.model_dump(), members=grp.members
            )
        except Exception:
            raise

    if grp is None:
        raise RuntimeError(
            f"Could not successfully validate {group} "
            "against any OME-Zarr group model.\n"
            "\n"
            "The following errors were encountered while trying to validate:\n\n"
            + "\n\n".join(
                f"{e[0].__module__}.{e[0].__name__}\n{type(e[1]).__name__}: {e[1]}"
                for e in errors
            )
        )
    return grp



"""
Test loading data from IDR.
"""

from contextlib import AbstractContextManager, nullcontext
from typing import Any

import pytest
import zarr

import ome_zarr_models.v05
from ome_zarr_models.exceptions import ValidationWarning


@pytest.mark.vcr
@pytest.mark.parametrize(
    ("url", "cls", "expected_warning"),
    [
        (
            "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.5/idr0066/ExpA_VIP_ASLM_on.zarr",
            ome_zarr_models.v05.Image,
            None,
        ),
        # The next dataset takes a long time (> 10 mins) to open, so comment out for now
        # ("https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.5/idr0010/76-45.ome.zarr",
        # ome_zarr_models.v05.HCS,
        # re.escape("'version' field not specified in plate metadata"))
    ],
)
def test_load_remote_data(
    url: str,
    cls: type[ome_zarr_models.v05.base.BaseGroupv05[Any]],
    expected_warning: str | None,
) -> None:
    ctx: AbstractContextManager[Any]
    if expected_warning is not None:
        ctx = pytest.warns(ValidationWarning, match=expected_warning)
    else:
        ctx = nullcontext()

    with ctx:
        grp = cls.from_zarr(zarr.open_group(url))
    assert isinstance(grp, ome_zarr_models.v05.Image)



from collections.abc import Sequence
from typing import Literal

from pydantic import JsonValue

from ome_zarr_models.base import BaseAttrs

__all__ = ["Axes", "Axis", "AxisType"]


AxisType = Literal["space", "time", "channel"]


class Axis(BaseAttrs):
    """
    Model for an element of `Multiscale.axes`.
    """

    # Explicitly name could be any JsonValue, but implicitly it must match Zarr array
    # dimension_names which limits it to str | None

    name: str | None
    type: str | None = None
    # Unit probably intended to be str, but the spec doesn't explicitly specify
    unit: str | JsonValue | None = None


Axes = Sequence[Axis]



from __future__ import annotations

import argparse
import sys
import warnings
from typing import TYPE_CHECKING, Literal

from ome_zarr_models import __version__, open_ome_zarr
from ome_zarr_models.exceptions import ValidationWarning

if TYPE_CHECKING:
    from zarr.storage import StoreLike


def main() -> None:
    """Main entry point for the ome-zarr-models CLI."""
    parser = argparse.ArgumentParser(
        prog="ome-zarr-models",
        description="OME-Zarr Models CLI",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"ome-zarr-models version {__version__}",
    )
    subparsers = parser.add_subparsers(
        dest="command", help="Available commands", metavar="COMMAND"
    )

    # validate sub-command
    validate_cmd = subparsers.add_parser("validate", help="Validate an OME-Zarr")
    validate_cmd.add_argument(
        "path", type=str, help="Path to OME-Zarr group to validate"
    )

    # info sub-command
    info_cmd = subparsers.add_parser(
        "info", help="Get information about an OME-Zarr group"
    )
    info_cmd.add_argument(
        "path", type=str, help="Path to OME-Zarr group to get information about"
    )

    args = parser.parse_args()

    # Execute the appropriate command
    match args.command:
        case "validate":
            validate(args.path)
        case "info":
            info(args.path)
        case None:
            parser.print_help()
            sys.exit(1)


def validate(path: StoreLike, version: Literal["0.4", "0.5"] | None = None) -> None:
    """Validate an OME-Zarr at the given path.

    Parameters
    ----------
    path : str
        Path to the OME-Zarr group to validate.  May be
    version : str | None, optional
        OME-Zarr version to validate against. If `None`, the version will be
        inferred from the metadata, by default `None`.

    Examples
    --------
    ```bash
    ome-zarr-models validate https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.5/idr0066/ExpD_chicken_embryo_MIP.ome.zarr
    ```
    """
    try:
        with warnings.catch_warnings(action="error", category=ValidationWarning):
            open_ome_zarr(path, version=version)
    except Exception as e:
        print(f"{e}\n")
        print(f"❌ Invalid OME-Zarr: {path}")
        sys.exit(1)
    print("✅ Valid OME-Zarr")


def info(path: StoreLike) -> None:
    """Print information about an OME-Zarr at the given path.

    Examples
    --------
    ```bash
    ome-zarr-models info https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.5/idr0066/ExpD_chicken_embryo_MIP.ome.zarr
    ome-zarr-models info https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.5/idr0062A/6001240_labels.zarr
    ome-zarr-models info https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.5/idr0083/9822152.zarr
    ome-zarr-models info https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.5/idr0010/76-45.ome.zarr
    ```
    """
    import zarr

    try:
        from rich import print
    except ImportError:
        from builtins import print

    try:
        group = zarr.open_group(path, mode="r")
        obj = open_ome_zarr(group)
    except Exception as e:
        print(f"{e}\n")
        print(f"❌ Invalid OME-Zarr: {path}")
        sys.exit(1)
    else:
        print(obj)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import TYPE_CHECKING, Self

from pydantic import Field

from ome_zarr_models._utils import _from_zarr_v2
from ome_zarr_models.base import BaseAttrsv2
from ome_zarr_models.v04.base import BaseGroupv04

if TYPE_CHECKING:
    import zarr

__all__ = ["Labels", "LabelsAttrs"]


class LabelsAttrs(BaseAttrsv2):
    """
    Attributes for an OME-Zarr labels dataset.
    """

    labels: list[str] = Field(
        ..., description="List of paths to labels arrays within a labels dataset."
    )


class Labels(BaseGroupv04[LabelsAttrs]):
    """
    An OME-Zarr labels dataset.
    """

    @classmethod
    def from_zarr(cls, group: zarr.Group) -> Self:  # type: ignore[override]
        """
        Create an OME-Zarr labels model from a `zarr.Group`.

        Parameters
        ----------
        group : zarr.Group
            A Zarr group that has valid OME-Zarr labels metadata.
        """
        return _from_zarr_v2(group, cls, LabelsAttrs)



from zarr.abc.store import Store

from ome_zarr_models._v06.well import Well, WellAttrs
from ome_zarr_models._v06.well_types import WellImage, WellMeta
from tests._v06.conftest import json_to_zarr_group


def test_well(store: Store) -> None:
    zarr_group = json_to_zarr_group(json_fname="well_example.json", store=store)
    ome_group = Well.from_zarr(zarr_group)
    assert ome_group.attributes.ome == WellAttrs(
        version="0.6",
        well=WellMeta(
            images=[
                WellImage(path="0", acquisition=1),
                WellImage(path="1", acquisition=1),
                WellImage(path="2", acquisition=2),
                WellImage(path="3", acquisition=2),
            ],
            version="0.6",
        ),
    )


def test_get_paths() -> None:
    well = WellMeta(
        images=[
            WellImage(path="0", acquisition=1),
            WellImage(path="1", acquisition=1),
            WellImage(path="2", acquisition=2),
            WellImage(path="3", acquisition=2),
        ],
        version="0.6",
    )

    assert well.get_acquisition_paths() == {1: ["0", "1"], 2: ["2", "3"]}



from ome_zarr_models.common.omero import Channel, Omero, Window
from tests.v04.conftest import read_in_json


def test_load_example_json() -> None:
    model = read_in_json(json_fname="omero_example.json", model_cls=Omero)

    assert model == Omero(
        channels=[
            Channel(
                color="0000FF",
                window=Window(max=65535.0, min=0.0, start=0.0, end=1500.0),
                active=True,
                coefficient=1,
                family="linear",
                inverted=False,
                label="LaminB1",
            )
        ],
        id=1,
        name="example.tif",
        version="0.4",
        rdefs={"defaultT": 0, "defaultZ": 118, "model": "color"},
    )



from typing import Literal

from pydantic import Field

import ome_zarr_models.common.well_types
from ome_zarr_models.common.well_types import WellImage

__all__ = ["WellImage", "WellMeta"]


class WellMeta(ome_zarr_models.common.well_types.WellMeta):
    """
    Metadata for a single well.
    """

    version: Literal["0.6"] | None = Field(
        None, description="Version of the well specification"
    )



from functools import partial
from typing import Literal

from tests import conftest

VERSION: Literal["0.6"] = "0.6"
json_to_dict = partial(conftest.json_to_dict, version=VERSION)
read_in_json = partial(conftest.read_in_json, version=VERSION)
json_to_zarr_group = partial(conftest.json_to_zarr_group, version=VERSION)
